part of pdf;

/// Represents the collection of pages in a [PdfDocument] .
class PdfPageCollection {
  //Constructor
  /// Initializes a new instance of the [PdfPageCollection] class.
  PdfPageCollection._(PdfDocument document, [PdfSection section]) {
    if (document == null) {
      throw ArgumentError.notNull('document');
    }
    _document = document;
    if (section != null) {
      _section = section;
    }
  }

  //Fields
  /// Represents the  method that executes on a PdfDocument when a new page is created.
  PageAddedCallback pageAdded;
  PdfDocument _document;
  int _count = 0;
  final Map<PdfPage, int> _pageCollectionIndex = <PdfPage, int>{};
  PdfSection _section;

  //Properties
  /// Gets the page by index
  PdfPage operator [](int index) => _returnValue(index);

  /// Gets the total number of the pages.
  int get count {
    if (_section == null) {
      return _countPages();
    } else {
      return _section._count;
    }
  }

  //Public methods
  /// Adds a new page.
  PdfPage add() {
    final PdfPage page = PdfPage();
    page._isNewPage = true;
    _addPage(page);
    page._isNewPage = false;
    return page;
  }

  /// Gets the index of the page.
  int indexOf(PdfPage page) {
    ArgumentError.checkNotNull(page, 'page');
    if (_section != null) {
      return _section._indexOf(page);
    } else {
      int index = -1;
      int numPages = 0;
      for (int i = 0; i < _document._sections.count; i++) {
        final PdfSection section = _document.sections[i];
        index = section._indexOf(page);
        if (index >= 0) {
          index += numPages;
          break;
        }
        numPages += section._count;
      }
      return index;
    }
  }

  //Implementation
  void _addPage(PdfPage page) {
    PdfSection section = _section != null ? _section : _getLastSection();
    if (_section == null) {
      if (section.pageSettings.orientation !=
          _document.pageSettings.orientation) {
        section = _document.sections.add();
        section.pageSettings.orientation = _document.pageSettings.orientation;
      }
      if (!_pageCollectionIndex.containsKey(page)) {
        _pageCollectionIndex[page] = _count++;
      }
    }
    section._add(page);
  }

  bool _contains(PdfPage page) {
    if (_section != null) {
      return _section._indexOf(page) != -1;
    } else {
      bool value = false;
      for (int i = 0; i < _document.sections.count; i++) {
        value |= _document.sections[i].pages._contains(page);
      }
      return value;
    }
  }

  PdfPage _returnValue(int index) {
    if (_section != null) {
      return _section._getPageByIndex(index);
    } else {
      if (index < 0 || index >= count) {
        throw ArgumentError.value('index', 'out of range');
      }

      PdfPage page;
      int sectionStartIndex = 0;
      int sectionCount = 0;
      int pageIndex = 0;
      for (int i = 0; i < _document.sections.count; i++) {
        final PdfSection section = _document.sections[i];
        sectionCount = section._count;
        pageIndex = index - sectionStartIndex;
        if (index >= sectionStartIndex && pageIndex < sectionCount) {
          page = section._getPageByIndex(pageIndex);
          break;
        }
        sectionStartIndex += sectionCount;
      }
      return page;
    }
  }

  PdfSection _getLastSection() {
    final PdfSectionCollection sectionCollection = _document.sections;
    if (sectionCollection._sections.isEmpty) {
      sectionCollection.add();
    }
    return sectionCollection[sectionCollection._sections.length - 1];
  }

  int _countPages() {
    final PdfSectionCollection sectionCollection = _document.sections;
    int count = 0;
    for (int i = 0; i < sectionCollection._sections.length; i++) {
      final PdfSection section = sectionCollection[i];
      count += section._count;
    }
    return count;
  }

  void _onPageAdded(PageAddedArgs args) {
    if (pageAdded != null) {
      pageAdded(this, args);
    }
  }

  void _remove(PdfPage page) {
    ArgumentError.checkNotNull(page);
    if (_section != null) {
      _section._remove(page);
    } else {
      for (int i = 0; i < _document.sections.count; i++) {
        if (_document.sections[i].pages._contains(page)) {
          _document.sections[i].pages._remove(page);
          break;
        }
      }
    }
  }
}
